@@ -7,6 +7,10 @@
  * (C) 2006 Brian Elliott Finley (finley@anl.gov)
  * (C) 2009-2010 Doug Engert (deengert@anl.gov)
  * (C) 2010 James Y Knight (foom@fuhm.net)
+ * (C) 2010-2013 Ken Dreyer <ktdreyer at ktdreyer.com>
+ * (C) 2012-2016 Mark Proehl <mark at mproehl.net>
+ * (C) 2012-2016 Olaf Flebbe <of at oflebbe.de>
+ * (C) 2013-2016 Daniel Kobras <d.kobras@science-computing.de>
  *
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -88,20 +92,47 @@ int generate_new_password(msktutil_flags *flags)
         }
     }
     close(fd);
-    VERBOSE(" Characters read from /dev/udandom = %d",chars_used);
+    VERBOSE(" Characters read from /dev/urandom = %d",chars_used);
     return 0;
 }
 
 
-int set_password(msktutil_flags *flags, int time)
+/* Try to set the the new Samba secret to <password>. */
+static int set_samba_secret(const std::string& password)
+{
+    VERBOSE("Setting samba machine trust account password");
+
+    FILE *pipe = popen("net changesecretpw -f -i", "w");
+    if (pipe == NULL) {
+        fprintf(stdout, "Error executing samba net command\n");
+        return 1;
+    }
+
+    size_t len = password.length();
+    if (fwrite(password.c_str(), sizeof(char), len, pipe) != len) {
+        fprintf(stdout, "Write error putting password to samba net command\n");
+        return 1;
+    }
+
+    int rc = pclose(pipe);
+    if (rc != 0) {
+        fprintf(stdout, "Setting samba secret failed with error code %d\n", rc);
+        return 1;
+    }
+
+    VERBOSE("Successfully set samba machine trust account password");
+
+    return 0;
+}
+
+
+int set_password(msktutil_flags *flags)
 {
     int ret;
     krb5_data resp_code_string;
     krb5_data resp_string;
     int response = 0;
     std::string old_pwdLastSet;
-    std::string current_pwdLastSet;
-
 
     // Zero out these data structures, because we attempt to free them below, and sometimes, upon
     // error conditions, the called API hasn't set them itself.
@@ -116,7 +147,7 @@ int set_password(msktutil_flags *flags, int time)
         VERBOSE("Attempting to reset computer's password");
     }
     if (flags->auth_type == AUTH_FROM_USER_CREDS) {
-        VERBOSE("Try change password using user's ticket cache\n");
+        VERBOSE("Try change password using user's ticket cache");
 
         KRB5CCache ccache(KRB5CCache::defaultName());
         KRB5Principal principal(flags->samAccountName);
@@ -150,37 +181,41 @@ int set_password(msktutil_flags *flags, int time)
         KRB5Creds creds;
         /* Use the machine's credentials */
         if (flags->auth_type == AUTH_FROM_SAM_KEYTAB ||
+            flags->auth_type == AUTH_FROM_SAM_UPPERCASE_KEYTAB ||
             flags->auth_type == AUTH_FROM_EXPLICIT_KEYTAB ||
             flags->auth_type == AUTH_FROM_HOSTNAME_KEYTAB) {
             std::string princ_name;
-            if (flags->auth_type == AUTH_FROM_SAM_KEYTAB)
+            if (flags->auth_type == AUTH_FROM_SAM_KEYTAB) {
                 princ_name = flags->samAccountName;
-            else if (flags->auth_type == AUTH_FROM_EXPLICIT_KEYTAB)
+            } else if (flags->auth_type == AUTH_FROM_SAM_UPPERCASE_KEYTAB) {
+                princ_name = flags->samAccountName_uppercase;
+            } else if (flags->auth_type == AUTH_FROM_EXPLICIT_KEYTAB) {
                 princ_name = flags->keytab_auth_princ;
-            else
+            } else {
                 princ_name = "host/" + flags->hostname;
-            VERBOSE("Try using keytab for %s to change password\n", princ_name.c_str());
+            }
+            VERBOSE("Try using keytab for %s to change password", princ_name.c_str());
 
             KRB5Keytab keytab(flags->keytab_readname);
             KRB5Principal principal(princ_name);
             KRB5Creds local_creds(principal, keytab, "kadmin/changepw");
             creds.move_from(local_creds);
         } else if (flags->auth_type == AUTH_FROM_PASSWORD) {
-            VERBOSE("Try using default password for %s to change password\n", flags->samAccountName.c_str());
+            VERBOSE("Try using default password for %s to change password", flags->samAccountName.c_str());
 
             KRB5Principal principal(flags->samAccountName);
             KRB5Creds local_creds(principal, create_default_machine_password(flags->samAccountName), "kadmin/changepw");
             creds.move_from(local_creds);
         } else if ((flags->auth_type == AUTH_FROM_SUPPLIED_PASSWORD) ||
                    (flags->auth_type == AUTH_FROM_SUPPLIED_EXPIRED_PASSWORD)) {
-            VERBOSE("Try using supplied password for %s to change password\n", flags->samAccountName.c_str());
+            VERBOSE("Try using supplied password for %s to change password", flags->samAccountName.c_str());
             KRB5Principal principal(flags->samAccountName);
             KRB5Creds local_creds(principal, flags->old_account_password, "kadmin/changepw");
             creds.move_from(local_creds);
         } else // shouldn't happen
             throw Exception("Error: unknown auth_type.");
 
-        if (flags->auth_type != AUTH_FROM_SUPPLIED_EXPIRED_PASSWORD ) {
+        if (flags->auth_type != AUTH_FROM_SUPPLIED_EXPIRED_PASSWORD) {
             old_pwdLastSet = ldap_get_pwdLastSet(flags);
         }
 
@@ -206,68 +241,13 @@ int set_password(msktutil_flags *flags, int time)
                 return ret;
             }
         }
-
     }
 
-    VERBOSE("Successfully set password, waiting for it to be reflected in LDAP.");
-    if (flags->auth_type == AUTH_FROM_SUPPLIED_EXPIRED_PASSWORD) {
+    VERBOSE("Successfully set password.");
 
-        VERBOSE("Warning: authenticated with expired password -- no way to verify the password change in LDAP.");
-
-    }  else {
-
-        /* Loop and wait for the account and password set to replicate */
-        for (int this_time = 0; ; this_time += 5) {
-            current_pwdLastSet = ldap_get_pwdLastSet(flags);
-            if (time + this_time >= 60) {
-                fprintf(stdout, "Password reset failed.\n");
-                return 1;
-            }
-            if (this_time >= 30) {
-                fprintf(stdout, "Re-attempting password reset for %s\n", flags->samAccountName.c_str());
-                return set_password(flags);
-            }
-            if (current_pwdLastSet.empty()) {
-                /* Account hasn't replicated yet */
-                fprintf(stdout, "Waiting for account replication (%d seconds past)\n", time + this_time);
-                sleep(5);
-            } else {
-                /* The account exists: we're waiting for the value to
-                 * change, indicating the password set worked */
-                if (current_pwdLastSet != old_pwdLastSet) {
-                    /* Password set has replicated successfully */
-                    VERBOSE("Successfully reset computer's password");
-                    break;
-                }
-                fprintf(stdout, "Waiting for password replication (%d seconds past)\n", time + this_time);
-                sleep(5);
-            }
-        }
+    if (!flags->set_samba_secret) {
+        return 0;
     }
 
-    if (flags->set_samba_secret) {
-        VERBOSE("Setting samba machine trust account password");
-
-        FILE *pipe = popen("net changesecretpw -f -i", "w");
-        if (pipe == NULL) {
-            fprintf(stdout, "Error executing samba net command\n");
-            return 1;
-        }
-
-        size_t len = flags->password.length();
-        if (fwrite(flags->password.c_str(), sizeof(char), len, pipe) != len) {
-            fprintf(stdout, "Write error putting password to samba net command\n");
-            return 1;
-        }
-
-        int rc = pclose(pipe);
-        if (rc != 0) {
-            fprintf(stdout, "Setting samba secret failed with error code %d\n", rc);
-            return 1;
-        }
-
-        VERBOSE("Successfully set samba machine trust account password");
-    }
-
-    return 0;
+    return set_samba_secret(flags->password);
 }
