@@ -7,6 +7,10 @@
  * (C) 2006 Brian Elliott Finley (finley@anl.gov)
  * (C) 2009-2010 Doug Engert (deengert@anl.gov)
  * (C) 2010 James Y Knight (foom@fuhm.net)
+ * (C) 2010-2013 Ken Dreyer <ktdreyer at ktdreyer.com>
+ * (C) 2012-2016 Mark Proehl <mark at mproehl.net>
+ * (C) 2012-2016 Olaf Flebbe <of at oflebbe.de>
+ * (C) 2013-2016 Daniel Kobras <d.kobras@science-computing.de>
  *
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -32,6 +36,7 @@
 #include "config.h"
 
 #include <stdio.h>
+#include <stdarg.h>
 #include <fcntl.h>
 #include <string.h>
 #include <unistd.h>
@@ -40,14 +45,12 @@
 #include <stdlib.h>
 #include <errno.h>
 #include <time.h>
+#include <limits.h>
 #include <netdb.h>
 #include <sys/socket.h>
 #include <sys/utsname.h>
 #include <ldap.h>
-#include <arpa/inet.h>
-#include <arpa/nameser.h>
-#include <resolv.h>
-
+#include <list>
 
 #ifdef HAVE_COM_ERR_H
 # ifdef COM_ERR_NEEDS_EXTERN_C
@@ -59,11 +62,7 @@
 # endif
 #endif
 #include <krb5.h>
-#ifdef HAVE_SASL_H
-#include <sasl.h>
-#else
-#include <sasl/sasl.h>
-#endif
+
 
 #include <stdexcept>
 #include <string>
@@ -77,8 +76,8 @@
 #define MAX_HOSTNAME_LEN                255
 #define MAX_TRIES                       10
 #define MAX_SAM_ACCOUNT_LEN             20
-#define MAX_DEF_MACH_PASS_LEN		14
-#define MAX_DOMAIN_CONTROLLERS		20
+#define MAX_DEF_MACH_PASS_LEN           14
+#define MAX_DOMAIN_CONTROLLERS          20
 
 
 #ifndef TMP_DIR
@@ -121,15 +120,11 @@
 #define KVNO_FAILURE                    -1
 #define KVNO_WIN_2000                   0
 
-/* LDAP Binding Attempts */
-#define ATTEMPT_SASL_PARAMS_TLS         0
-#define ATTEMPT_SASL_NO_PARAMS_TLS      1
-#define ATTEMPT_SASL_NO_TLS             2
-
 /* Ways we can authenticate */
 enum auth_types {
     AUTH_NONE = 0,
     AUTH_FROM_SAM_KEYTAB,
+    AUTH_FROM_SAM_UPPERCASE_KEYTAB,
     AUTH_FROM_HOSTNAME_KEYTAB,
     AUTH_FROM_PASSWORD,
     AUTH_FROM_USER_CREDS,
@@ -154,10 +149,12 @@ enum msktutil_mode {
     MODE_UPDATE,
     MODE_AUTO_UPDATE,
     MODE_FLUSH,
+    MODE_CLEANUP,
     MODE_PRECREATE
 };
 
-struct msktutil_flags {
+class msktutil_flags {
+public:
     std::string keytab_file;
     std::string keytab_writename;
     std::string keytab_readname;
@@ -171,23 +168,25 @@ struct msktutil_flags {
     std::string base_dn;
     std::string samAccountName;
     std::string samAccountName_nodollar;
+    std::string samAccountName_uppercase;
     std::string password;
+    bool password_from_cmdline;
     std::string userPrincipalName;
     std::string unicodePwd;
     std::string old_account_password;
     std::string site;
-    std::auto_ptr<LDAPConnection> ldap;
-
+    LDAPConnection* ldap;
     std::string ad_computerDn;
     std::string ad_dnsHostName;
-    std::string ad_userPrincipal;
     std::vector<std::string> ad_principals;
 
     bool set_description;
     bool set_userPrincipalName;
     bool no_reverse_lookups;
+    bool no_canonical_name;
     bool server_behind_nat;
     bool set_samba_secret;
+    bool check_replication;
 
     msktutil_val dont_expire_password;
     msktutil_val no_pac;
@@ -203,45 +202,65 @@ struct msktutil_flags {
     bool use_service_account;
     bool allow_weak_crypto;
     bool password_expired;
+    int auto_update_interval;
+    krb5_kvno kvno;
+    int cleanup_days;
+    int cleanup_enctype;
     msktutil_flags();
+
+private:
+    msktutil_flags operator=(const msktutil_flags& other);
+    msktutil_flags(const msktutil_flags& other);
     ~msktutil_flags();
 };
 
-struct msktutil_exec {
+class msktutil_exec {
+public:
     msktutil_mode mode;
     std::vector<std::string> add_principals;
     std::vector<std::string> remove_principals;
-    msktutil_flags *flags;
 
     msktutil_exec();
+
     ~msktutil_exec();
+    void set_mode(msktutil_mode mode);
 };
 
-struct msktutil_dcdata {
-    char srvname[NS_MAXDNAME];
-    unsigned int priority;
-    unsigned int weight;
-    unsigned int port; 
+
+class Globals {
+    msktutil_flags* _flags;
+    msktutil_exec* _exec;
+    static Globals *instance;
+
+public:
+    static Globals* get();
+    static msktutil_flags *flags() {
+        return Globals::get()->_flags;
+    }
+    static msktutil_exec *exec() {
+        return Globals::get()->_exec;
+    }
+    void set_supportedEncryptionTypes(char * value);
 };
 
 /* Prototypes */
 extern std::string create_default_machine_password(const std::string &samaccountname);
 extern void ldap_cleanup(msktutil_flags *);
 extern void init_password(msktutil_flags *);
-extern std::string get_default_hostname();
+extern std::string get_default_hostname(bool no_canonical_name = false);
 extern void get_default_keytab(msktutil_flags *);
 extern void get_default_ou(msktutil_flags *);
-extern std::auto_ptr<LDAPConnection> ldap_connect(const std::string &server,
-                                                  bool no_reverse_lookups = false,
-                                                  int try_tls=ATTEMPT_SASL_PARAMS_TLS);
+
 extern void ldap_get_base_dn(msktutil_flags *);
-extern std::string complete_hostname(const std::string &);
+extern std::string complete_hostname(const std::string &,
+                                     bool no_canonical_name = false);
 extern std::string get_short_hostname(msktutil_flags *);
 extern int flush_keytab(msktutil_flags *);
+extern void cleanup_keytab(msktutil_flags *);
 extern void update_keytab(msktutil_flags *);
-extern void add_principal_keytab(const std::string &, krb5_kvno kvno, msktutil_flags *);
+extern void add_principal_keytab(const std::string &, msktutil_flags *);
 extern int ldap_flush_principals(msktutil_flags *);
-extern int set_password(msktutil_flags *, int time = 0);
+extern int set_password(msktutil_flags *);
 extern krb5_kvno ldap_get_kvno(msktutil_flags *);
 extern std::string ldap_get_pwdLastSet(msktutil_flags *);
 extern std::vector<std::string> ldap_list_principals(msktutil_flags *);
@@ -251,7 +270,8 @@ extern std::string get_dc_host(const std::string &realm_name, const std::string
                                const bool);
 extern std::string get_user_principal();
 extern std::string get_host_os();
-extern void ldap_check_account(msktutil_flags *);
+extern bool ldap_check_account(msktutil_flags *);
+extern void ldap_create_account(msktutil_flags *);
 extern void create_fake_krb5_conf(msktutil_flags *);
 extern void remove_fake_krb5_conf();
 extern void remove_ccache();
@@ -263,7 +283,6 @@ int generate_new_password(msktutil_flags *flags);
 #define VERBOSE(text...) if (g_verbose) { fprintf(stdout, " -- %s: ", __FUNCTION__); fprintf(stdout, ## text); fprintf(stdout, "\n"); }
 
 
-#define VERBOSEldap(text...) if (g_verbose > 1) { fprintf(stderr, " ###### %s: ", __FUNCTION__); fprintf(stderr, ## text); fprintf(stderr, "\n"); }
 #endif
 
 // printf into a C++ string.
@@ -318,26 +337,7 @@ class LDAPException : public Exception
 
 
 #include "krb5wrap.h"
+#include "ldapconnection.h"
 
 
 
-class LDAPConnection {
-public: //fixme
-    LDAP *m_ldap;
-
-public:
-    LDAPConnection(const std::string &server);
-
-    void set_option(int option, const void *invalue);
-    void get_option(int option, void *outvalue);
-    void start_tls(LDAPControl **serverctrls=NULL, LDAPControl **clientctrls=NULL);
-
-    void search(LDAPMessage **mesg_p,
-                const std::string &base_dn, int scope, const std::string &filter, char *attrs[],
-                int attrsonly=0, LDAPControl **serverctrls=NULL, LDAPControl **clientctrls=NULL,
-                struct timeval *timeout=NULL, int sizelimit=-1);
-
-    std::string get_one_val(LDAPMessage *mesg, char *name);
-    std::vector<std::string> get_all_vals(LDAPMessage *mesg, char *name);
-    ~LDAPConnection();
-};
